@@ -1,7 +1,8 @@
 from fontTools import ttLib
-from fontTools.ttLib import ttFont, tables
+from fontTools import subset
+from fontTools.ttLib import ttFont, tables, newTable, TTFont
 from fontTools.ttLib.tables.TupleVariation import TupleVariation
-from fontTools.ufoLib.filenames import userNameToFileName
+import ufoLib2.objects as ufol
 from lxml import etree
 from ast import literal_eval
 try:
@@ -12,13 +13,14 @@ try:
     from .ygridfit import ygridfit_parse, ygridfit_parse_obj
 except ImportError:
     from ygridfit import ygridfit_parse, ygridfit_parse_obj
-from tempfile import mkstemp
+from tempfile import mkstemp, NamedTemporaryFile, SpooledTemporaryFile
 import sys
 import os
 import argparse
 import array
 import re
 import pkg_resources
+import unicodedata
 
 #      This file is part of xgridfit, version 3.
 #      Licensed under the Apache License, Version 2.0.
@@ -30,14 +32,14 @@ neutral_instructions = [ 'IUP', 'RDTG', 'ROFF', 'RTDG', 'RTG', 'RTHG', 'RUTG',
 
 # How many numbers popped by each instruction (-1 means a variable number)
 
-pop_instructions = { 'ALIGNPTS': 2, 'ALIGNRP': -1, 'IP': -1, 'MDAP': 1,
-                     'MIAP': 2, 'MIRP': 2, 'MDRP': 1, 'SHP': -1, 'SLOOP': 1,
-                     'SRP0': 1, 'SRP1': 1, 'SRP2': 1, 'CALL': 1, 'SFVTL': 2,
-                     'SPVTL': 2, 'SDPVTL': 2, 'ISECT': 5, 'MSIRP': 2, 'SCFS': 2,
-                     'SCVTCI': 1, 'SFVFS': 2, 'SHC': 1, 'SHZ': 1, 'SMD': 1,
-                     'SPVFS': 2, 'SROUND': 1, 'SSW': 1, 'SSWCI': 1, 'SZP0': 1,
-                     'SZP1': 1, 'SZP2': 1, 'SZPS': 1, 'UTP': 1, 'WCVTF': 2,
-                     'WCVTP': 2, 'WS': 2 }
+pop_instructions = { 'ALIGNPTS': 2, 'ALIGNRP': -1, 'IP': -1,   'MDAP': 1,
+                     'MIAP': 2,     'MIRP': 2,     'MDRP': 1,  'SHP': -1,  'SLOOP': 1,
+                     'SRP0': 1,     'SRP1': 1,     'SRP2': 1,  'CALL': 1,  'SFVTL': 2,
+                     'SPVTL': 2,    'SDPVTL': 2,   'ISECT': 5, 'MSIRP': 2, 'SCFS': 2,
+                     'SCVTCI': 1,   'SFVFS': 2,    'SHC': 1,   'SHZ': 1,   'SMD': 1,
+                     'SPVFS': 2,    'SROUND': 1,   'SSW': 1,   'SSWCI': 1, 'SZP0': 1,
+                     'SZP1': 1,     'SZP2': 1,     'SZPS': 1,  'UTP': 1,   'WCVTF': 2,
+                     'WCVTP': 2,    'WS': 2 }
 
 byte_push_instructions = [ 'PUSHB', 'NPUSHB' ]
 
@@ -47,6 +49,12 @@ quietcount = 0
 
 coordinateFuzz = 1
 
+ufo_mode = False
+
+class xgfGlyphObject:
+    def __init__(self):
+        self.lib = None
+
 def get_file_path(fn):
     return pkg_resources.resource_filename(__name__, fn)
 
@@ -71,6 +79,9 @@ def wipe_font(fo):
     except Exception:
         pass
 
+def wipe_font_ufo(ufo):
+    pass
+
 def install_glyph_program(nm, fo, asm):
     """ name of the glyph, the font object, the instructions in one big string."""
     try:
@@ -86,6 +97,24 @@ def install_glyph_program(nm, fo, asm):
         print("name is correct and that it exists in the font.")
         sys.exit(1)
 
+def install_glyph_program_ufo(ufo, nm, asm):
+    try:
+        global maxInstructions
+        p = tables.ttProgram.Program()
+        p.fromAssembly(asm)
+        b = len(p.getBytecode())
+        if b > maxInstructions:
+            maxInstructions = b
+        glyph = ufo[nm]
+        if not "public.truetype.instructions" in glyph.lib:
+            glyph.lib["public.truetype.instructions"] = {}
+        glyph.lib["public.truetype.instructions"]["formatVersion"] = "1"
+        glyph.lib["public.truetype.instructions"]["assembly"] = str(asm)
+    except Exception as e:
+        print("error in install_glyph_program_ufo:")
+        print(e.args)
+        print(e)
+
 def is_number(s):
     try:
         int(s)
@@ -191,6 +220,26 @@ def install_cvt(fo, cvs, base):
         fo['cvt '] = ttFont.newTable('cvt ')
         setattr(fo['cvt '],'values',array.array('h', cvs))
 
+def install_cvt_ufo(ufo, cvs, base):
+    cvt_dict = {}
+    if base > 0:
+        cvt_index = base
+        try:
+            old_cvt_dict = ufo.lib["public.truetype.instructions"]["controlValue"]
+        except Exception:
+            old_cvt_dict = {}
+    else:
+        cvt_index = 0
+    for c in cvs:
+        cvt_dict[str(cvt_index)] = c
+        cvt_index += 1
+    if base > 0:
+        cvt_dict.update(old_cvt_dict)
+    if not "public.truetype.instructions" in ufo.lib:
+        ufo.lib["public.truetype.instructions"] = {}
+    ufo.lib["public.truetype.instructions"]["formatVersion"] = "1"
+    ufo.lib["public.truetype.instructions"]["controlValue"] = cvt_dict
+
 def install_functions(fo, fns, base):
     """If base greater than zero, we append our own functions to those already
     in the font. If not, generate a new fpgm table."""
@@ -204,6 +253,23 @@ def install_functions(fo, fns, base):
         fo['fpgm'].program = tables.ttProgram.Program()
         fo['fpgm'].program.fromAssembly(str(fns))
 
+def install_functions_ufo(ufo, fns, base):
+    oldfpgm = ""
+    if base > 0:
+        try:
+            oldfpgm = "\n".join(ufo.lib["public.truetype.instructions"]["fontProgram"])
+        except Exception as e:
+            print("error in install_functions_ufo:")
+            print(e.args)
+            print(e)
+            pass
+    newfpgm = oldfpgm + str(fns)
+    if not "public.truetype.instructions" in ufo.lib:
+        ufo.lib["public.truetype.instructions"] = {}
+    ufo.lib["public.truetype.instructions"]["formatVersion"] = "1"
+    ufo.lib["public.truetype.instructions"]["fontProgram"] = newfpgm
+
+
 def install_prep(fo, pr, keepold, replace):
     """If keepold is True, we append our own prep table to the existing one.
     It is up to the Xgridfit programmer to make sure that no conflicts arise
@@ -219,6 +285,39 @@ def install_prep(fo, pr, keepold, replace):
         fo['prep'].program = tables.ttProgram.Program()
         fo['prep'].program.fromAssembly(str(pr))
 
+def install_prep_ufo(ufo, pr, keepold, replace):
+    oldprep = ""
+    if keepold and not replace:
+        try:
+            oldprep = "\n".join(ufo.lib["public.truetype.instructions"]["controlValueProgram"])
+        except Exception as e:
+            print("error in install_prep_ufo:")
+            print(e.args)
+            print(e)
+            # pass
+    newprep = oldprep + str(pr)
+    if not "public.truetype.instructions" in ufo.lib:
+        ufo.lib["public.truetype.instructions"] = {}
+    ufo.lib["public.truetype.instructions"]["formatVersion"] = "1"
+    ufo.lib["public.truetype.instructions"]["controlValueProgram"] = newprep
+
+def install_properties_ufo(ufo, d):
+    # keys can be:
+    # maxFunctionDefs: int
+    # maxInstructionDefs: int
+    # maxStackElements: int
+    # maxStorage: int
+    # maxSizeOfInstructions: int
+    # maxTwilightPoints: int
+    # maxZones: int
+    if not "public.truetype.instructions" in ufo.lib:
+        ufo.lib["public.truetype.instructions"] = {}
+    ufo.lib["public.truetype.instructions"]["formatVersion"] = "1"
+    k = d.keys()
+    for kk in k:
+        ufo.lib["public.truetype.instructions"][kk] = d[kk]
+    
+
 # Compare two dictionaries. It would be easier to do an equality test, but
 # I can't be sure two otherwise matching dictionaries aren't in a different
 # order.
@@ -520,12 +619,19 @@ def validate(f, syntax, noval):
         schema = etree.RelaxNG(etree.parse(schemapath))
         schema.assertValid(f)
 
-def compile_all(font, yaml, new_file_name):
+def compile_all(font: ttFont, yaml: dict, new_file_name: str) -> list:
+    global ufo_mode
     xgffile = ygridfit_parse_obj(yaml)
     ns = {"xgf": "http://xgridfit.sourceforge.net/Xgridfit2",
           "xi": "http://www.w3.org/2001/XInclude",
           "xsl": "http://www.w3.org/1999/XSL/Transform"}
-    thisFont = ttLib.TTFont(font)
+    split_fn = os.path.splitext(new_file_name)
+    extension = split_fn[1]
+    ufo_mode = (extension == ".ufo")
+    ufo = None
+    if ufo_mode:
+        ufo = ufol.Font.open(new_file_name)
+    thisFont = font
     functionBase = 0     # Offset to account for functions in existing font
     cvtBase      = 0     # Offset to account for CVs in existing font
     storageBase  = 0     # Offset to account for storage in existing font
@@ -534,7 +640,6 @@ def compile_all(font, yaml, new_file_name):
     wipe_font(thisFont)
     xslfile = etree.parse(get_file_path("XSL/xgridfit-ft.xsl"))
     etransform = etree.XSLT(xslfile)
-    # glyph_list = [gname]
     glyph_list = str(etransform(xgffile, **{"get-glyph-list": "'yes'"}))
     glyph_list = list(glyph_list.split(" "))
     coordinateIndex = make_coordinate_index(glyph_list, thisFont)
@@ -543,25 +648,42 @@ def compile_all(font, yaml, new_file_name):
     safe_calls = literal_eval(str(safe_calls))
     cvt_list = str(etransform(xgffile, **{"get-cvt-list": "'yes'"}))
     cvt_list = literal_eval("[" + cvt_list + "]")
-    install_cvt(thisFont, cvt_list, cvtBase)
+    if ufo_mode:
+        install_cvt_ufo(ufo, cvt_list, cvtBase)
+    else:
+        install_cvt(thisFont, cvt_list, cvtBase)
     cvar_count = len(xgffile.xpath("/xgf:xgridfit/xgf:cvar", namespaces=ns))
-    if cvar_count > 0:
+    if cvar_count > 0 and not ufo_mode:
         tuple_store = literal_eval(str(etransform(xgffile, **{"get-cvar": "'yes'"})))
         install_cvar(thisFont, tuple_store, False, cvtBase)
     predef_functions = int(xslfile.xpath("/xsl:stylesheet/xsl:variable[@name='predefined-functions']",
                                          namespaces=ns)[0].attrib['select'])
     maxFunction = etransform(xgffile, **{"function-count": "'yes'"})
     maxFunction = int(maxFunction) + predef_functions + functionBase
-    thisFont['maxp'].maxSizeOfInstructions = maxInstructions + 50
-    thisFont['maxp'].maxTwilightPoints = twilightBase + 25
-    thisFont['maxp'].maxStorage = storageBase + 64
-    thisFont['maxp'].maxStackElements = maxStack
-    thisFont['maxp'].maxFunctionDefs = maxFunction
-    thisFont['head'].flags |= 0b0000000000001000
+    if ufo_mode:
+        d = {"maxSizeOfInstructions": maxInstructions + 50,
+             "maxTwilightPoints": twilightBase + 25,
+             "maxStorage": 64,
+             "maxStackElements": maxStack,
+             "maxFunctionDefs": maxFunction}
+        install_properties_ufo(ufo, d)
+    else:
+        thisFont['maxp'].maxSizeOfInstructions = maxInstructions + 50
+        thisFont['maxp'].maxTwilightPoints = twilightBase + 25
+        thisFont['maxp'].maxStorage = storageBase + 64
+        thisFont['maxp'].maxStackElements = maxStack
+        thisFont['maxp'].maxFunctionDefs = maxFunction
+        thisFont['head'].flags |= 0b0000000000001000
     fpgm_code = etransform(xgffile, **{"fpgm-only": "'yes'"})
-    install_functions(thisFont, fpgm_code, functionBase)
+    if ufo_mode:
+        install_functions_ufo(ufo, fpgm_code, functionBase)
+    else:
+        install_functions(thisFont, fpgm_code, functionBase)
     prep_code = etransform(xgffile, **{"prep-only": "'yes'"})
-    install_prep(thisFont, prep_code, False, True)
+    if ufo_mode:
+        install_prep_ufo(ufo, prep_code, False, True)
+    else:
+        install_prep(thisFont, prep_code, False, True)
     failed_glyph_list = []
     for g in glyph_list:
         try:
@@ -571,60 +693,74 @@ def compile_all(font, yaml, new_file_name):
             #    glyph_args['init_graphics'] = "'" + initgraphics + "'"
             g_inst = etransform(xgffile, **glyph_args)
             g_inst_final = compact_instructions(str(g_inst), safe_calls)
-            install_glyph_program(g, thisFont, g_inst_final)
+            if ufo_mode:
+                install_glyph_program_ufo(ufo, g, g_inst_final)
+            else:
+                install_glyph_program(g, thisFont, g_inst_final)
         except Exception as e:
+            print("error in compile_all:")
+            print(e.args)
             print(e)
             for entry in etransform.error_log:
-                print('message from line %s, col %s: %s' % (entry.line, entry.column, entry.message))
+                if quietcount < 3:
+                    print('message from line %s, col %s: %s' % (entry.line, entry.column, entry.message))
             # sys.exit(1)
             failed_glyph_list.append(g)
-    with thisFont as f:
-        f.save(new_file_name, 1)
+    if ufo_mode:
+        ufo.save()
+    else:
+        with thisFont as f:
+            f.save(new_file_name, 1)
     return failed_glyph_list
 
-def compile_one(font, yaml, gname):
-    """ A quick and dirty function to support the GUI. It compiles code for a
-        single glyph and generates the font with that code (and supporting
-        stuff--cvt, cvar, prep, fpgm, maxp). This can then be used to display
-        a preview of the glyph.
+def compile_list(font: ttFont, yaml: dict, gname: str) -> tuple:
     """
-    xgffile = ygridfit_parse_obj(yaml)
+        Compile code either for a single glyph or a list of glyphs. This will
+        generate a hinted font (as a SpooledTemporaryFile) with *only* those glyphs
+        in it. It returns this open file, plus a name:gid dict and a list of
+        glyphs that failed to compile.
+    """
+    # No output to console.
+    global quietcount
+    quietcount = 3
+    if type(gname) is list:
+        xgffile = ygridfit_parse_obj(yaml, glyph_list=gname)
+    else:
+        xgffile = ygridfit_parse_obj(yaml, single_glyph=gname)
     ns = {"xgf": "http://xgridfit.sourceforge.net/Xgridfit2",
           "xi": "http://www.w3.org/2001/XInclude",
           "xsl": "http://www.w3.org/1999/XSL/Transform"}
-    thisFont = ttLib.TTFont(font)
-    functionBase = 0     # Offset to account for functions in existing font
-    cvtBase      = 0     # Offset to account for CVs in existing font
-    storageBase  = 0     # Offset to account for storage in existing font
-    maxStack     = 256   # Our (generous) default stack size
-    twilightBase = 0     # Offset to account for twilight space in existing font
+    thisFont = font
     wipe_font(thisFont)
     xslfile = etree.parse(get_file_path("XSL/xgridfit-ft.xsl"))
     etransform = etree.XSLT(xslfile)
-    glyph_list = [gname]
+    if type(gname) is str:
+        glyph_list = [gname]
+    else:
+        glyph_list = gname
     coordinateIndex = make_coordinate_index(glyph_list, thisFont)
     coordinates_to_points(glyph_list, xgffile, coordinateIndex, ns)
     safe_calls = etransform(xgffile, **{"stack-safe-list": "'yes'"})
     safe_calls = literal_eval(str(safe_calls))
     cvt_list = str(etransform(xgffile, **{"get-cvt-list": "'yes'"}))
     cvt_list = literal_eval("[" + cvt_list + "]")
-    install_cvt(thisFont, cvt_list, cvtBase)
+    install_cvt(thisFont, cvt_list, 0)
     cvar_count = len(xgffile.xpath("/xgf:xgridfit/xgf:cvar", namespaces=ns))
     if cvar_count > 0:
         tuple_store = literal_eval(str(etransform(xgffile, **{"get-cvar": "'yes'"})))
-        install_cvar(thisFont, tuple_store, False, cvtBase)
+        install_cvar(thisFont, tuple_store, False, 0)
     predef_functions = int(xslfile.xpath("/xsl:stylesheet/xsl:variable[@name='predefined-functions']",
                                          namespaces=ns)[0].attrib['select'])
     maxFunction = etransform(xgffile, **{"function-count": "'yes'"})
-    maxFunction = int(maxFunction) + predef_functions + functionBase
+    maxFunction = int(maxFunction) + predef_functions
     thisFont['maxp'].maxSizeOfInstructions = maxInstructions + 50
-    thisFont['maxp'].maxTwilightPoints = twilightBase + 25
-    thisFont['maxp'].maxStorage = storageBase + 64
-    thisFont['maxp'].maxStackElements = maxStack
+    thisFont['maxp'].maxTwilightPoints = 25
+    thisFont['maxp'].maxStorage = 64
+    thisFont['maxp'].maxStackElements = 256
     thisFont['maxp'].maxFunctionDefs = maxFunction
     thisFont['head'].flags |= 0b0000000000001000
     fpgm_code = etransform(xgffile, **{"fpgm-only": "'yes'"})
-    install_functions(thisFont, fpgm_code, functionBase)
+    install_functions(thisFont, fpgm_code, 0)
     prep_code = etransform(xgffile, **{"prep-only": "'yes'"})
     install_prep(thisFont, prep_code, False, True)
     failed_glyph_list = []
@@ -635,18 +771,38 @@ def compile_one(font, yaml, gname):
             #if initgraphics:
             #    glyph_args['init_graphics'] = "'" + initgraphics + "'"
             g_inst = etransform(xgffile, **glyph_args)
-            g_inst_final = compact_instructions(str(g_inst), safe_calls)
-            install_glyph_program(g, thisFont, g_inst_final)
+            # g_inst_final = compact_instructions(str(g_inst), safe_calls)
+            # install_glyph_program(g, thisFont, g_inst_final)
+            # g_inst_final = str(g_inst)
+            # install_glyph_program(g, thisFont, g_inst_final)
+            install_glyph_program(g, thisFont, str(g_inst))
         except Exception as e:
             print(e)
             for entry in etransform.error_log:
-                print('message from line %s, col %s: %s' % (entry.line, entry.column, entry.message))
+                if quietcount < 3:
+                    print('message from line %s, col %s: %s' % (entry.line, entry.column, entry.message))
             failed_glyph_list.append(g)
-    tmp_file_name = mkstemp(suffix=".ttf")[1]
-    with thisFont as f:
-        f.save(tmp_file_name, 1)
-    # print("Temp font saved to " + str(tmp_file_name))
-    return tmp_file_name, failed_glyph_list
+    try:
+        # Grab the font, subset it with just the one glyph we're
+        # previewing, and write it to a spooled temporary file.
+        tf = SpooledTemporaryFile(max_size=1000000, mode='b')
+        options = subset.Options(glyph_names=True)
+        options.layout_features = []
+        subsetter = subset.Subsetter(options)
+        subsetter.populate(glyphs=glyph_list)
+        subsetter.subset(thisFont)
+        glyph_id = {}
+        for g in glyph_list:
+            glyph_id[g] = thisFont.getGlyphID(g)
+        # temp diagnostic:
+        # thisFont.save("temp_font.ttf", 1)
+        thisFont.save(tf, 1)
+        tf.seek(0)
+    except Exception as e:
+        print(e)
+    # Return: a handle to the temp file, the index of the target glyph, a list of
+    # glyphs for which complilation failed
+    return tf, glyph_id, failed_glyph_list
 
 
 
@@ -721,7 +877,8 @@ def main():
     if cfuzz > 1:
         coordinateFuzz = cfuzz
 
-    if os.path.splitext(inputfile)[1] == ".yaml":
+    source_ext = os.path.splitext(inputfile)[1]
+    if source_ext == ".yaml":
         if quietcount < 1:
             print("Converting yaml source to Xgridfit")
         xgffile = ygridfit_parse(inputfile)
@@ -940,11 +1097,14 @@ def main():
         except Exception as e:
             print(e)
             for entry in etransform.error_log:
-                print('message from line %s, col %s: %s' % (entry.line, entry.column, entry.message))
-            sys.exit(1)
+                if quietcount < 3:
+                    print('message from line %s, col %s: %s' % (entry.line, entry.column, entry.message))
+            # sys.exit(1)
+            continue
         # Two lines added for (possible) debugging
         for entry in etransform.error_log:
-            print('message from line %s, col %s: %s' % (entry.line, entry.column, entry.message))
+            if quietcount < 3:
+                print('message from line %s, col %s: %s' % (entry.line, entry.column, entry.message))
         if nocompact or g in no_compact_list:
             g_inst_final = str(g_inst)
         else:
